# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, InstanceBooleanError, InvalidBranchNameError, \
    DuplicateBranchError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches, output, uids, sizes, symbols, paint_branch, paint_header, \
    paint_layer


def arguments(branch, name, package=False):
    """
    - Gets the arguments and parses them to the 'duplicate_branch' function.
    """
    duplicate_branch(branch, name, package)


def duplicate_branch(argument_branch, argument_name, argument_package):
    """
    Errors:
      - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If the name value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If the package value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
      - If the name value contains a character that is not a letter or number, then it raises an
        'InvalidBranchError' error.

    Branches list loop:
      Errors:
        - If the branch is already in the 'branches' dictionary, then it raises a 'DuplicateBranchError' error.

    Branches list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

    - If the value of the 'argument_package' variable is true, then update the values of all required dictionaries with
      the values of the given package argument.
    - If the value of the 'argument_package' variable is false, then update the values of all required dictionaries with
      the default values for a new branch.
    """
    if not isinstance(argument_branch, str):
        raise InstanceStringError("branch", argument_branch)

    if not isinstance(argument_name, str):
        raise InstanceStringError("name", argument_name)

    if not isinstance(argument_package, bool):
        raise InstanceBooleanError("package", argument_package)

    if not argument_name.isalnum():
        raise InvalidBranchNameError(argument_name)

    for branches_branch in list(branches):
        if argument_name.lower() == branches_branch.lower():
            raise DuplicateBranchError(argument_name)

    error = 0
    for branches_branch in list(branches):
        if argument_branch.lower() == branches_branch.lower():
            error = error + 1

            branches.update({argument_name: branches[branches_branch]})
            output.update({argument_name: []})

            if argument_package:
                uids.update({argument_name: uids[branches_branch]})
                sizes.update({argument_name: sizes[branches_branch]})
                symbols.update({argument_name: symbols[branches_branch]})
                paint_branch.update({argument_name: paint_branch[branches_branch]})
                paint_header.update({argument_name: paint_header[branches_branch]})
                paint_layer.update({argument_name: paint_layer[branches_branch]})
            else:
                uids.update({argument_name: []})
                sizes.update({argument_name: 0})
                symbols.update({argument_name: {"line": "┃", "split": "┣━", "end": "┗━"}})
                paint_branch.update({argument_name: []})
                paint_header.update({argument_name: []})
                paint_layer.update({argument_name: {}})

    if error == 0:
        if argument_branch:
            raise NotExistingBranchError(argument_branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceIntegerError, InvalidSizeError, InstanceStringError, \
    NotExistingBranchError
from br4nch.utility.utility_librarian import branches, symbols, sizes


def arguments(branch, size):
    """
    - Gets the arguments and parses them to the 'set_size' function.
    """
    set_size(branch, size)


def set_size(argument_branch, argument_size):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the size value is not an instance of a integer, then it raises an 'InstanceIntegerError' error.
      - If the size value is smaller than '0' or bigger than '20', then it raises an 'InvalidSizeError' error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch value is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      Branches list loop:
        - If the branch is in the 'branches' dictionary, it checks if the rule variables are default and if so, the
          horizontal line is added times the value of the variable 'argument_size'.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_size, int):
        raise InstanceIntegerError("size", argument_size)

    if argument_size < 0 or argument_size > 20:
        raise InvalidSizeError

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                if symbols[branches_branch]["split"] == "┣━":
                    symbols[branches_branch]["split"] = "┣━" + "━" * argument_size
                if symbols[branches_branch]["end"] == "┗━":
                    symbols[branches_branch]["end"] = "┗━" + "━" * argument_size

                sizes.update({branches_branch: argument_size})

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.create.create_branch import arguments as branch
from br4nch.create.create_layer import arguments as layer



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import os

from br4nch.utility.utility_handler import InstanceBooleanError, InstanceStringError, NotExistingDirectoryError, \
    NotExistingBranchError
from br4nch.utility.utility_librarian import branches, uids, sizes, symbols, paint_branch, paint_header, paint_layer
from br4nch.utility.utility_printer import printer


def arguments(branch, package=False, beautify=True, directory=""):
    """
    - Gets the arguments and parses them to the 'export_branch' function.
    """
    export_branch(branch, package, beautify, directory)


def export_branch(argument_branch, argument_package, argument_beautify, argument_directory):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the package value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
      - If the beautify value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
      - If the directory value is not an instance of a string, then it raises an 'InstanceStringError' error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If and the branch value is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      Branches list loop:
        - If the branch is in the 'branches' dictionary, then it is checked whether the variable 'argument_package' is
          equal to true, if the variable is equal to true, then a dictionary is created with the values of all necessary
          dictionaries.
        - Then it is checked whether the variable 'argument_directory' is equal to false, the 'printer' function is
          called with the 'export_branch' action that prints the given 'branches_branch' dictionary and the
          'branches_branch' package dictionary. If the 'argument_beautify' variable is false, the given dictionary(s)
          are not represented with a branch structure.

        If argument directory is equal to true:
          If the given directory exists:
            - Creates the custom br4nch directory if it does not exist in the given directory argument.
            - Appends the value of the branch in the 'branches' directory to a custom file with the id line for a
              exported branch file in the custom br4nch folder.
            - If there is a value in the 'argument_package' argument, then adds the values of all required directories
              to a custom file with the id line for a exported package file in the custom br4nch folder.

          - If the given directory argument does not exist, it will throw a 'NotExistingDirectoryError' error.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_package, bool):
        raise InstanceBooleanError("package", argument_package)

    if not isinstance(argument_beautify, bool):
        raise InstanceBooleanError("beautify", argument_beautify)

    if not isinstance(argument_directory, str):
        raise InstanceStringError("directory", argument_directory)

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                export_package = {}

                if argument_package:
                    export_package = {branches_branch: [uids[branches_branch], sizes[branches_branch],
                                                        symbols[branches_branch], paint_branch[branches_branch],
                                                        paint_header[branches_branch], paint_layer[branches_branch]]}

                if not argument_directory:
                    printer("export_branch", [branches_branch, {branches_branch: branches[branches_branch]},
                                              export_package, argument_beautify])
                else:
                    if os.path.isdir(argument_directory):
                        if not os.path.isdir(argument_directory + "/br4nch-" + branches_branch):
                            os.mkdir(argument_directory + "/br4nch-" + branches_branch)

                        with open(argument_directory + "/br4nch-" + branches_branch + "/branch-" + branches_branch,
                                  'w', encoding='utf-8') as file:
                            file.write("id=:br4nch-branch:\n")
                            file.write(str({branches_branch: branches[branches_branch]}))

                        if argument_package:
                            with open(argument_directory + "/br4nch-" + branches_branch + "/package-" + branches_branch,
                                      'w', encoding='utf-8') as file:
                                file.write("id=:br4nch-package:\n")
                                file.write(str(export_package))
                    else:
                        raise NotExistingDirectoryError(argument_directory)

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceBooleanError, InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches, paint_branch, paint_header, paint_layer
from br4nch.utility.utility_builder import Builder
from br4nch.utility.utility_printer import printer


def arguments(branch, delete=False):
    """
    - Gets the arguments and parses them to the 'display_branch' function.
    """
    display_branch(branch, delete)


def display_branch(argument_branch, argument_delete):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the delete value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      Branches list loop:
        Paint validator:
          - If there is a value in the value of the key branch of the 'paint_branch' dictionary, then the first element
            in the list 'total' is updated with the value + 1.
          - If there is a value in the value of the key branch of the 'paint_header' dictionary, then the first element
            in the list 'total' is updated with the value + 1.
          - Then the 'paint_layer' is looped through with the value of the key branch. For each 'layer' value, it is
            checked whether that layer/value has a value in the 'paint_layer' dictionary with the key branch. If there
            is no value in the 'layer' key, then the 'no_paint' variable is added with the value + 1. If the value of
            the 'no_paint' variable is equal to the value of the length of the 'paint_layer' dictionary with the key
            branch, then the first element in the list 'total' is updated with the value + 1.
          - If the first element in the 'total' list is equal to '3', the 'colored' variable becomes 'false'. Otherwise
            the 'colored' variable becomes 'true'.

        - Calls the 'Builder' class with the current value of branch as argument.
        - Calls the 'printer' function with the 'display_branch' action and the current 'branches_branch' and
          'argument_delete' argument.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_delete, bool):
        raise InstanceBooleanError("delete", argument_delete)

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                total = [0]
                if not paint_branch[branches_branch]:
                    total[0] = total[0] + 1

                if not paint_header[branches_branch]:
                    total[0] = total[0] + 1

                no_paint = 0
                for layer in paint_layer[branches_branch]:
                    if not paint_layer[branches_branch][layer]:
                        no_paint = no_paint + 1

                if no_paint == len(paint_layer[branches_branch]):
                    total[0] = total[0] + 1

                if total[0] == 3:
                    colored = False
                else:
                    colored = True

                Builder(branches_branch, colored)
                printer("display_branch", [branches_branch], argument_delete)

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches, paint_branch
from br4nch.utility.utility_handler import InstanceStringError, NotExistingBranchError


def arguments(branch):
    """
    - Gets the arguments and parses them to the 'reset_paint_branch' function.
    """
    reset_paint_branch(branch)


def reset_paint_branch(argument_branch):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Branches list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      - If the branch is in the 'branches' dictionary, then the value of the current branch key in the 'paint_branch'
        dictionary is updated to an empty list.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                paint_branch.update({branches_branch: []})

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import os
import ast

from br4nch.utility.utility_handler import NotExistingBranchFileError, InvalidBranchFileError, \
    InstanceDictionaryError, NotExistingPackageFileError, InvalidPackageFileError, DuplicateBranchError
from br4nch.utility.utility_librarian import branches, output, uids, sizes, symbols, paint_branch, paint_header, \
    paint_layer


def arguments(branch, package=""):
    """
    - Gets the arguments and parses them to the 'load_branch' function.
    """
    load_branch(branch, package)


def load_branch(argument_branch, argument_package):
    """
    Load branch:
      Errors:
        - Raises an 'NotExistingBranchFileError' error if the instance of the 'argument_branch' variable is a string and
          the given directory does not exists.
        - If given branch file does not have the required branch id tag or the length of the total lines is less than
          '2', then it raises an 'InvalidBranchFileError' error.
        - If the 'argument_branch' is not a instance of a string and/or the given directory does not exists, the content
          of the branch file will be assigned to the 'argument_branch' variable. then it raises an
          'InstanceDictionaryError' error.

      - If the 'argument_branch' is instance of a string and the given directory exists, the content of the branch file
        will be assigned to the 'argument_branch' variable.

    Load package:
      Errors:
        - Raises an 'NotExistingPackageFileError' error if there is value in the the 'argument_package' variable and is
          instance of string and the given directory does not exists.
        - If given package file does not have the required package id tag or the length of the total lines is less than
          '2', then it raises an 'InvalidPackageFileError' error.
        - If there is a value in the 'argument_package' variable and is not a instance of a string and/or the given
          directory does not exists, then it raises an 'InstanceDictionaryError' error.

      - If there is a value in the 'argument_package' variable and instance of a string and the given directory exists,
        the content of the file package file will be assigned to the 'argument_package' variable.

    Branches list loop:
      Errors:
        - If the branch is already in the 'branches' dictionary, then it raises a 'DuplicateBranchError' error.

    - If the value of the 'argument_package' variable is true, then update the values of all required dictionaries with
      the values of the given package argument.
    - If the value of the 'argument_package' variable is false, then update the values of all required dictionaries with
      the default values for a new branch.
    """
    if isinstance(argument_branch, str) and not os.path.isfile(argument_branch):
        raise NotExistingBranchFileError(argument_branch)
    elif isinstance(argument_branch, str) and os.path.isfile(argument_branch):
        with open(argument_branch, 'r', encoding="utf8") as file:
            file = file.readlines()

            if len(file) < 2:
                raise InvalidBranchFileError(argument_branch)

            if str(file[0][:-1]) != "id=:br4nch-branch:":
                raise InvalidBranchFileError(argument_branch)

            argument_branch = ast.literal_eval(file[1])
    else:
        if not isinstance(argument_branch, dict):
            raise InstanceDictionaryError("branch", argument_branch)

    if argument_package and isinstance(argument_package, str) and not os.path.isfile(argument_package):
        raise NotExistingPackageFileError(argument_package)
    elif argument_package and isinstance(argument_package, str) and os.path.isfile(argument_package):
        with open(argument_package, 'r', encoding="utf8") as file:
            file = file.readlines()

            if len(file) < 2:
                raise InvalidPackageFileError(argument_package)

            if str(file[0][:-1]) != "id=:br4nch-package:":
                raise InvalidPackageFileError(argument_package)

            argument_package = ast.literal_eval(file[1])
    else:
        if argument_package and not isinstance(argument_package, dict):
            raise InstanceDictionaryError("package", argument_package)

    if argument_branch:
        for branches_branch in list(branches):
            if list(argument_branch)[0].lower() == branches_branch.lower():
                raise DuplicateBranchError(list(argument_branch)[0])

        branches.update({list(argument_branch)[0]: list(argument_branch.values())[0]})
        output.update({list(argument_branch)[0]: []})

        if argument_package and list(argument_package)[0] == list(argument_branch)[0] and len(list(argument_package.values())[0]) == 6:
            uids.update({list(argument_branch)[0]: list(argument_package.values())[0][0]})
            sizes.update({list(argument_branch)[0]: list(argument_package.values())[0][1]})
            symbols.update({list(argument_branch)[0]: list(argument_package.values())[0][2]})
            paint_branch.update({list(argument_branch)[0]: list(argument_package.values())[0][3]})
            paint_header.update({list(argument_branch)[0]: list(argument_package.values())[0][4]})
            paint_layer.update({list(argument_branch)[0]: list(argument_package.values())[0][5]})
        else:
            uids.update({list(argument_branch)[0]: []})
            sizes.update({list(argument_branch)[0]: 0})
            symbols.update({list(argument_branch)[0]: {"line": "┃", "split": "┣━", "end": "┗━"}})
            paint_branch.update({list(argument_branch)[0]: []})
            paint_header.update({list(argument_branch)[0]: []})
            paint_layer.update({list(argument_branch)[0]: {}})



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

"""
Dictionaries:
  - branches: Stores the given header and layers in this dictionary for each branch.
  - output: Stores the final result of the 'builder' class in this dictionary for each branch.
  - uids: Stores all generated UIDS in this dictionary for each branch.
  - sizes: Stores all given sizes in this dictionary for each branch.
  - symbols: Stores all given 'line', 'split' and 'end' strings in this dictionary for each branch.
  - paint_branch: Stores all given paint for the branch in this dictionary for each branch.
  - paint_header: Stores all given header paint in this dictionary for each branch.
  - paint_layer: Stores all given paint for the layers in this dictionary for each branch.
"""
branches = {}
output = {}
uids = {}
sizes = {}
symbols = {}
paint_branch = {}
paint_header = {}
paint_layer = {}



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceBooleanError, InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches
from br4nch.utility.utility_printer import printer


def arguments(branch, layer, sensitive=False, beautify=True):
    """
    - Gets the arguments and parses them to the 'DisplayLayer' class.
    """
    DisplayLayer(branch, layer, sensitive, beautify)


class DisplayLayer:
    def __init__(self, argument_branch, argument_layer, argument_sensitive, argument_beautify):
        """
        - Gets the arguments and parses them to the 'build_position_structure' function.
        """
        self.build_position_structure(argument_branch, argument_layer, argument_sensitive, argument_beautify)

    def build_position_structure(self, argument_branch, argument_layer, argument_sensitive, argument_beautify):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given pos argument is not an instance of a list, then the branch argument will be set as a list.

        Errors:
          - If the sensitive value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
          - If the beautify value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            Errors:
              - If the 'loop_layer' value is not an instance of a string, then it raises an 'InstanceStringError' error.

            - Calls the 'elevator' function to calculate each level/height of each layer and stores the result in the
              levels list. Then a '0' is added to the 'levels' list so that the 'IndexError' error can be avoided.
            - If the branch is in the 'branches' dictionary, it will loop with all positions in the returned list of the
              'format_position' function. And calls the 'display_layer' function for each loop with the built position
              structure of the 'position' as argument
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_layer, list):
            argument_layer = [argument_layer]

        if not isinstance(argument_sensitive, bool):
            raise InstanceBooleanError("sensitive", argument_sensitive)

        if not isinstance(argument_beautify, bool):
            raise InstanceBooleanError("beautify", argument_beautify)

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    levels = [0]
                    self.elevator(levels, branches[branches_branch][list(branches[branches_branch])[0]])
                    levels.append(0)

                    for layer in argument_layer:
                        if not isinstance(layer, str):
                            raise InstanceStringError("layer", layer)

                        self.display_layer(branches_branch, layer, argument_sensitive, argument_beautify, levels, [0],
                                           branches[branches_branch][list(branches[branches_branch])[0]])

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def elevator(self, levels, value, pos=0):
        """
        Value dictionary loop:
          - Loops through each "height" of the value dictionary and adds the value of the 'pos' variable to the 'levels'
            list.
          - Then the 'elevator' function is called again with the current value of the 'value' dictionary.
        """
        for value in value.values():
            levels.append(pos)

            self.elevator(levels, value, pos + 1)

    def display_layer(self, branch, loop_layer, argument_sensitive, argument_beautify, levels, trace, value,
                      position_structure=""):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable and the first element of the 'trace' list is
            added with plus '1'.

          Position structure:
            - Then it is checked whether the "level"/"height" of the current value of the loop is equal to or smaller
              than the previous "level"/"height" value of the loop. If the instantaneous value is equal or less, the
              last position and dot in the 'position_structure' variable is removed.
            - Then the current value of the 'position_structure' variable is added with the value of 'count' separated
              by a dot to the 'position_structure' variable.

          - If the 'argument_sensitive' is 'True', then is checked if the current value of 'layer' is equal to the value
            of 'loop_layer'. If the 'argument_sensitive' is 'False', then is checked if the current value of 'layer' in
            lower text is equal to the value of 'loop_layer' in lower text.
            - If the values are equal, then the 'printer' function is called and a package is supplied with all the
              values in it that are needed. If the 'argument_beautify' variable is false, the given positions are not
              represented with a branch structure.

          - Checks whether the 'value' variable has a value. If there is a value, then the 'display_layer' function is
            called again with the current values of 'value', 'trace' and 'levels' as arguments.
        """
        count = 0

        for layer, value in value.items():
            count = count + 1

            trace[0] = trace[0] + 1

            if levels[trace[0]] <= levels[trace[0] - 1]:
                position_structure = position_structure[:-2]

            position_structure = position_structure + "." + str(count)

            if argument_sensitive:
                if layer[:-15] == loop_layer:
                    printer("display_layer", [branch, layer[:-15], position_structure[1:], argument_beautify])
            else:
                if layer[:-15].lower() == loop_layer.lower():
                    printer("display_layer", [branch, layer[:-15], position_structure[1:], argument_beautify])

            if value:
                self.display_layer(branch, loop_layer, argument_sensitive, argument_beautify, levels, trace, value,
                                   position_structure)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import uuid

from br4nch.utility.utility_librarian import uids


def generate_uid(branch, length=10):
    """
    - Generates the uid.

    While loop:
      - If the uid exists in the current branch uids list, a new one is generated.
      - If the uid does not exist in the current branch uids list, then the uid is added to the branch uids list.

    - Returns the generated uid.
    """

    uid = str(uuid.uuid4()).replace("-", "")[0:length]

    while True:
        if uid in uids[branch]:
            return generate_uid(branch, length)
        else:
            uids[branch].append(uid)
            break

    return ":uid=" + uid



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches, paint_layer
from br4nch.utility.utility_handler import InstanceStringError, NotExistingBranchError, NotExistingPaintError
from br4nch.utility.utility_positioner import format_position


def arguments(branch, position, paint):
    """
    - Gets the arguments and parses them to the 'SetPaintLayer' class.
    """
    SetPaintLayer(branch, position, paint)


class SetPaintLayer:
    def __init__(self, argument_branch, argument_position, argument_paint):
        """
        - Gets the arguments and parses them to the 'build_position_structure' function.
        """
        self.build_position_structure(argument_branch, argument_position, argument_paint)

    def build_position_structure(self, argument_branch, argument_position, argument_paint):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given position argument is not an instance of a list, then the position argument will be set as a
            list.
          - If the given paint argument is not an instance of a list, then the paint argument will be set as a list.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch value is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            - Loops through the obtained values of the current value of the 'position' variable and adds all values of
              the variable to the 'position_structure' variable.
            - Calls the 'set_paint_layer' function with the current value of the whole branch dictionary, the 'position'
              and 'argument_paint' variables as arguments.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_position, list):
            argument_position = [argument_position]

        if not isinstance(argument_paint, list):
            argument_paint = [argument_paint]

        for length in range(len(argument_position)):
            argument_position[length] = argument_position[length].split(".")

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    for position in format_position(branches_branch, argument_position.copy()):
                        self.set_paint_layer(branches_branch, argument_paint, position,
                                             branches[branches_branch][list(branches[branches_branch])[0]])

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def set_paint_layer(self, branch, paint_package, position, value):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'position':
            Errors:
              - If the paint element argument is not an instance of a string, then it raises an 'InstanceStringError'
                error.

            - If the length of the 'position' list is equal to '1' then it check if the 'paint' exists in the given
              list. If it does not exists, it raises a 'NotExistingPaintError' error. If it does exists in the list, it
              will add the 'paint_package' list to the 'paint_layer' list.
            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'position' list will be removed and the 'set_paint_layer' function will be
              called again with the new values of the 'value' variable as argument.
        """
        count = 0

        for layer, value in value.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    for paint in paint_package:
                        if not isinstance(paint, str):
                            raise InstanceStringError("paint", paint)

                        if paint.lower() not in ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white",
                                                 "bold", "underline", "reversing"]:
                            raise NotExistingPaintError(paint)

                    paint_layer[branch].update({layer: paint_package})
                else:
                    if value:
                        position.pop(0)
                        return self.set_paint_layer(branch, paint_package, position, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches, output, uids, sizes, symbols, paint_branch, paint_header, paint_layer


def printer(action, package, delete=False):
    """
    display_branch:
      - Prints the output in the given 'output' list and clears the 'output' list. If the 'delete' variable is 'True',
        then all necessary dictionaries containing the key value of the value of 'branch' will be deleted.
    display_pos:
      - Prints the results of the found layers of the given positions. If the 'package[3]'/beautify element is 'False',
        then the output will be printed without modified print.
    display_layer:
      - Prints the results of the found positions of the given layers. If the 'package[3]'/beautify element is 'False',
        then the output will be printed without modified print.
    export_branch:
      - If the 'package[2]' has a value, then it prints the exported branch dictionary with package. If the
        'package[3]'/beautify element is 'False', then the output will be printed without modified print.
      - If the 'package[2]' does not have a value, then it prints the exported branch dictionary without package. If the
        'package[3]'/beautify element is 'False', then the output will be printed without modified print.
    """
    branch = package[0]

    if action == "display_branch":
        for line in output[branch]:
            print(line)

        output[branch].clear()

        if delete:
            del branches[branch]
            del output[branch]
            del sizes[branch]
            del symbols[branch]
            del paint_branch[branch]
            del paint_header[branch]
            del paint_layer[branch]
            del uids[branch]

    if action == "display_pos":
        if package[3]:
            print("Get Position Result" + ":\n" + "┗━ Branch: " + branch + "\n" + " " * 3 + "┗━ Layer: "
                  + package[1].replace("\n", " ") + "\n" + " " * 6 + "┗━ Position: " + package[2])
        else:
            print(package[1])

    if action == "display_layer":
        if package[3]:
            print("Get Layer Result" + ":\n" + "┗━ Branch: " + branch + "\n" + " " * 3 + "┗━ Layer: "
                  + package[1].replace("\n", " ") + "\n" + " " * 6 + "┗━ Position: " + package[2])
        else:
            print(package[2])

    if action == "export_branch":
        if package[2]:
            if package[3]:
                print("Export Result" + ":\n" + "┗━ Branch: " + branch + "\n" + " " * 3 + "┣━ Export: "
                      + str(package[1]) + "\n" + " " * 3 + "┗━ Package: " + str(package[2]))
            else:
                print(str(package[1]))
                print(str(package[2]))
        else:
            if package[3]:
                print("Export Result" + ":\n" + "┗━ Branch: " + branch + "\n" + " " * 3 + "┗━ Export: "
                      + str(package[1]))
            else:
                print(str(package[1]))



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

class InstanceStringError(Exception):
    """
    - If the given argument value is not an instance of a string, then it raises an 'InstanceStringError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, argument, value):
        self.argument = argument
        self.value = value

    def __str__(self):
        return "The " + str(self.argument) + " argument: '" + str(self.value) + "' must be an instance of a 'str' and" \
                                                                                " not '" \
               + str(type(self.value).__name__) + "'."


class InstanceBooleanError(Exception):
    """
    - If the given argument value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, argument, value):
        self.argument = argument
        self.value = value

    def __str__(self):
        return "The " + str(self.argument) + " argument: '" + str(self.value) + "' must be an instance of a 'bool' " \
                                                                               "and not '" \
               + str(type(self.value).__name__) + "'."


class InstanceIntegerError(Exception):
    """
    - If the given argument value is not an instance of a integer, then it raises an 'InstanceIntegerError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, argument, value):
        self.argument = argument
        self.value = value

    def __str__(self):
        return "The " + str(self.argument) + " argument: '" + str(self.value) + "' must be an instance of a 'int' and" \
                                                                                " not '" \
               + str(type(self.value).__name__) + "'."


class InstanceDictionaryError(Exception):
    """
    - If the given argument value is not an instance of a dictionary, then it raises an 'InstanceDictionaryError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, argument, value):
        self.argument = argument
        self.value = value

    def __str__(self):
        return "The " + str(self.argument) + " argument: '" + str(self.value) + "' must be an instance of a 'dict' " \
                                                                                "and not '" \
               + str(type(self.value).__name__) + "'."


class InvalidBranchNameError(Exception):
    """
    - If the branch value contains a character that is not a letter or number, then it raises an
      'InvalidBranchNameError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "The branch name: '" + self.value + "' is not valid. Only numbers and/or letters may be used to add a" \
                                                   "branch."


class InvalidPositionError(Exception):
    """
    - If the pos argument is not equal to a number and/or operator, then it raises an 'InvalidPositionError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, argument, position):
        self.argument = argument
        self.position = position

    def __str__(self):
        return "The position: '" + self.position + "' is not valid. Only numbers and operators may be used to add a " \
                                                   "position to the " + self.argument + " argument."


class InvalidSizeError(Exception):
    """
    - If the size value is smaller than '0' or bigger than '20', then it raises an 'InvalidSizeError' error.
    """
    __module__ = Exception.__module__

    def __str__(self):
        return "The sizes that can be used is '0-20'."


class InvalidBranchFileError(Exception):
    """
    - If given branch file does not have the required branch id tag or the length of the total lines is less than '2',
      then it raises an 'InvalidBranchFileError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, file):
        self.file = file

    def __str__(self):
        return "The file: '" + self.file + "' is not valid as a branch file."


class InvalidPackageFileError(Exception):
    """
    - If given package file does not have the required branch id tag or the length of the total lines is less than '2',
      then it raises an 'InvalidPackageFileError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, file):
        self.file = file

    def __str__(self):
        return "The file: '" + self.file + "' is not valid as a package file."


class NotExistingBranchError(Exception):
    """
    - If the branch is not in the 'branches' dictionary, it raises a 'NotExistingBranchError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, branch):
        self.branch = branch

    def __str__(self):
        return "The branch: '" + str(self.branch) + "' does not exists."


class NotExistingDirectoryError(Exception):
    """
    - If the given directory argument does not exist, it will throw a 'NotExistingDirectoryError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, directory):
        self.directory = directory

    def __str__(self):
        return "The directory: '" + str(self.directory) + "' does not exist."


class NotExistingBranchFileError(Exception):
    """
    - Raises an 'NotExistingBranchFileError' error if the instance of the 'argument_branch' variable is a string and the
      given directory does not exists.
    """
    __module__ = Exception.__module__

    def __init__(self, file):
        self.file = file

    def __str__(self):
        return "The branch file: '" + str(self.file) + "' does not exist."


class NotExistingPackageFileError(Exception):
    """
    - Raises an 'NotExistingPackageFileError' error if there is value in the the 'argument_package' variable and is
      instance of string and the given directory does not exists.
    """
    __module__ = Exception.__module__

    def __init__(self, file):
        self.file = file

    def __str__(self):
        return "The package file: '" + str(self.file) + "' does not exist."


class NotExistingPaintError(Exception):
    """
    - If the current value of the position in the 'paint' list is not equal to any of the values in the 'paint_id' list,
      then it raises a 'NotExistingPaintError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, paint):
        self.paint = paint

    def __str__(self):
        return "The paint: '" + str(self.paint) + "' does not exists."


class MaximumPaintSlotsError(Exception):
    """
    - If the length of the 'paint' list is bigger than '4', then it raises a 'MaximumPaintSlotsError' error.
    """
    __module__ = Exception.__module__

    def __str__(self):
        return "The maximum paint slots that can be used is '4'."


class DuplicateBranchError(Exception):
    """
    - If the branch value is already in the 'branches' dictionary, then it raises a 'DuplicateBranchError' error.
    """
    __module__ = Exception.__module__

    def __init__(self, branch):
        self.branch = branch

    def __str__(self):
        return "The branch: '" + str(self.branch) + "' already exists."


class RequiredSymbolChangeError(Exception):
    """
    - If there is no value in the 'line', 'split' and 'end' arguments, then it raises an 'RequiredSymbolChangeError'
      error.
    """
    __module__ = Exception.__module__

    def __str__(self):
        return "Change at least one of the given symbols: 'line', 'split' or 'end'."



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import os

from br4nch.utility.utility_librarian import output, uids, sizes, symbols, paint_branch, paint_header, paint_layer, \
    branches
from br4nch.utility.utility_generator import generate_uid
from br4nch.utility.utility_handler import InstanceStringError, InstanceBooleanError, NotExistingDirectoryError, \
    InvalidBranchNameError, DuplicateBranchError


def arguments(branch, directory, header="", include="", exclude="", unused=True, folder_priority=True):
    """
    - Gets the arguments and parses them to the 'LoadFolder' class.
    """
    LoadFolder(branch, directory, header, include, exclude, unused, folder_priority)


class LoadFolder:
    def __init__(self, argument_branch, argument_directory, argument_header, argument_include, argument_exclude,
                 argument_unused, argument_folder_priority):
        """
        - Gets the arguments and parses them to the 'load_folder' function.
        """
        self.load_folder(argument_branch, argument_directory, argument_header, argument_include, argument_exclude,
                         argument_unused, argument_folder_priority)

    def load_folder(self, argument_branch, argument_directory, argument_header, argument_include, argument_exclude,
                    argument_unused, argument_folder_priority):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given include argument is not an instance of a list, then the include argument will be set as a list.
          - If the given exclude argument is not an instance of a list, then the exclude argument will be set as a list.
          - Raises an 'NotExistingDirectoryError' error if the given 'argument_branch' value directory does not exists.

        Errors:
          - If the directory value is not an instance of a string, then it raises an 'InstanceStringError' error.
          - If the header value is not an instance of a string, then it raises an 'InstanceStringError' error.
          - If the unused value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
          - If the folder_priority value is not an instance of a boolean, then it raises an 'InstanceBooleanError'
            error.

        - If the 'argument_header' argument does not have a value, then the root 'argument_directory' argument will be
          used as header value.

        - If the 'argument_folder_priority' variable is True, then it changes 'True' to 'False' and if the
          'argument_folder_priority' variable is False, then it changes 'False' to 'True'.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch value contains a character that is not a letter or number, then it raises an
              'InvalidBranchError' error.

          Branches list loop:
            Errors:
              - If the branch is already in the 'branches' dictionary, then it raises a 'DuplicateBranchError' error.

          - Loops through the given folder and adds all files to the list of 'paths'.

          - The 'path_length' variable is used to get the total length of all folders in the 'argument_directory'
            variable using the '.split' function. The 'path_length' variable is then used to slice all folders except
            the last one in the 'argument_directory' variable.

          Paths length loop:
            If 'argument_exclude':
              argument_exclude list loop:
                - If the first character in the 'extension' variable is equal to '.', then a '.' will be appended to the
                  'extension' value.
                - If the current extension from the 'path' value is equal to the 'extension' value, then it deletes the
                  file from the current 'path' value.

            If 'argument_exclude':
              argument_include list loop:
                - If the first character in the 'extension' variable is equal to '.', then a '.' will be appended to the
                  'extension' value.
                - If the current extension from the 'path' value is not equal to the 'extension' value, then it deletes
                  the file from the current 'path' value.

          Paths list loop:
            Path number split loop:
              - If the 'number' value '-1' is bigger or equal to '0', then it will slice all numbers bigger then the
                current number in the 'path' split list.
              - Calls the 'create_structure' function to calculate and add the current path part to the 'structure'
                dictionary.

          - Then it will add the current 'branch' key in all the mandatory dictionaries with the needed value.
          - It will then call the 'generate_layers_uids' function that adds a uid to all layers in the 'structure'
            dictionary.
          - Then it will add the current 'structure' dictionary with the 'branch' value as key to the 'branches'
            dictionary.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_include, list):
            argument_include = [argument_include]

        if not isinstance(argument_exclude, list):
            argument_exclude = [argument_exclude]

        if not isinstance(argument_directory, str):
            raise InstanceStringError("directory", argument_directory)

        if not isinstance(argument_header, str):
            raise InstanceStringError("header", argument_header)

        if not isinstance(argument_unused, bool):
            raise InstanceBooleanError("unused", argument_unused)

        if not isinstance(argument_folder_priority, bool):
            raise InstanceBooleanError("folder_priority", argument_folder_priority)

        if not os.path.isdir(argument_directory):
            raise NotExistingDirectoryError(argument_directory)

        if not argument_header:
            argument_header = argument_directory

        if argument_folder_priority:
            argument_folder_priority = False
        else:
            argument_folder_priority = True

        for branch in argument_branch:
            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            if not branch.isalnum():
                raise InvalidBranchNameError(branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    raise DuplicateBranchError(branch)

            structure = {argument_header: {}}
            paths = []

            for root, dirs, files in os.walk(argument_directory, topdown=argument_folder_priority):
                paths.append(root.replace("\\", "/"))

                for file in files:
                    paths.append(root.replace("\\", "/") + "/" + file)

            path_length = len(argument_directory.replace("\\", "/").split("/"))

            for number in range(len(paths)):
                if argument_exclude:
                    for extension in argument_exclude:
                        if extension:
                            if not extension[0] == ".":
                                extension = "." + extension

                            if paths[number][-len(extension):] == extension:
                                if paths[number] in paths:
                                    if not argument_unused:
                                        paths[number] = "/"
                                    else:
                                        path = ""
                                        if "." in paths[number].split("/")[-1]:
                                            for folder in paths[number].split("/")[:-1]:
                                                path = path + "/" + folder

                                            paths[number] = path[1:]

                if argument_include:
                    for extension in argument_include:
                        if extension:
                            if not extension[0] == ".":
                                extension = "." + extension

                            if paths[number][-len(extension):] != extension:
                                if paths[number] in paths:
                                    if not argument_unused:
                                        paths[number] = "/"
                                    else:
                                        path = ""
                                        if "." in paths[number].split("/")[-1]:
                                            for folder in paths[number].split("/")[:-1]:
                                                path = path + "/" + folder

                                            paths[number] = path[1:]

            for path in paths:
                for number in range(len(path.split("/")[path_length:])):
                    previous_file = []

                    if number - 1 >= 0:
                        previous_file = path.split("/")[path_length:number - len(path.split("/")[path_length:])]

                    self.create_structure(path.split("/")[path_length:][number], [argument_header] + previous_file,
                                          structure)

            output.update({branch: []})
            uids.update({branch: []})
            sizes.update({branch: 0})
            symbols.update({branch: {"line": "┃", "split": "┣━", "end": "┗━"}})
            paint_branch.update({branch: []})
            paint_header.update({branch: []})
            paint_layer.update({branch: {}})

            self.generate_layers_uids(branch, structure[list(structure)[0]])

            branches.update({branch: structure})

    def create_structure(self, file, previous_file, value):
        """
        Value dictionary loop:
          - If the first element in the 'previous_file' list is equal to the 'layer' value, then the first element
            in the 'previous_file' list is popped.
          - If there is no value in the 'previous_file' list and the 'file' value is not in the current value of
            'value', then the 'file' variable is added to the 'value' dictionary as key and an empty dictionary as
            value to the key.
          - If there is an value in the 'previous_file' list and/or the 'file' value exists in the current value of
            'value', then the 'create_structure' function is called with the new 'value' value.
        """
        for layer, value in value.items():
            if previous_file and layer == previous_file[0]:
                previous_file.pop(0)

                if not previous_file and file not in value:
                    return value.update({file: {}})
                else:
                    return self.create_structure(file, previous_file, value)

    def generate_layers_uids(self, branch, value):
        """
        Value dictionary loop:
          - Generates a new uid for the current 'layer' value and replaces it in all required dictionaries.
          - If there is a value in the 'value' variable, then the 'generate_layers_uids' function is called again with
            the new 'value' value as argument.
        """
        previous_value = value

        for layer, value in value.copy().items():
            uid_layer = layer + generate_uid(branch)

            previous_value[uid_layer] = previous_value.pop(layer)
            paint_layer[branch].update({uid_layer: []})

            if value:
                self.generate_layers_uids(branch, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import copy

from br4nch.utility.utility_librarian import branches, output, uids, sizes, symbols, paint_branch, paint_header, \
    paint_layer
from br4nch.utility.utility_handler import InstanceStringError, InstanceIntegerError, InvalidSizeError, \
    NotExistingBranchError
from br4nch.utility.utility_generator import generate_uid
from br4nch.display.display_branch import display_branch


def arguments(branch, size=0, line="", split="", end=""):
    """
    - Gets the arguments and parses them to the 'DisplayAssist' class.
    """
    DisplayAssist(branch, size, line, split, end)


class DisplayAssist:
    def __init__(self, argument_branch, argument_size, argument_line, argument_split, argument_end):
        """
        - Gets the arguments and parses them to the 'display_assist' function.
        """
        self.display_assist(argument_branch, argument_size, argument_line, argument_split, argument_end)

    def display_assist(self, argument_branch, argument_size, argument_line, argument_split, argument_end):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

        Errors:
          - If the size value is not an instance of a integer, then it raises an 'InstanceIntegerError' error.
          - If the line value is not an instance of a string, then it raises an 'InstanceStringError' error.
          - If the split value is not an instance of a string, then it raises an 'InstanceStringError' error.
          - If the end value is not an instance of a string, then it raises an 'InstanceStringError' error.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            - Calls the 'elevator' function to calculate each level/height of each layer and stores the result in the
              levels list. Then a '0' is added to the 'levels' list so that the 'IndexError' error can be avoided and
              after that the new value from the 'levels' list is returned.
            - If the branch is in the 'branches' dictionary, then a new name is generated for the value of the
              'branches_branch' variable. If the name already exists in the 'branches' dictionary, the loop starts again
              until a unique branch name is generated that does not yet exist in the 'branches' dictionary.
            - It then creates a deep copy of the current value of the 'branches_branch' value of the 'branches'
              directory and it is added as value to the 'branches' dictionary with the unique branch name as the key.
            - If there is no value in any of the 'line', 'split' or 'end' arguments, the value that is currently active
              of the missing arguments is used.
            - Then the unique branch name is added to the corresponding dictionaries with the default values as the
              value. The 'argument_size' variable will be updated to the 'sizes' dictionary.
            - Calls the function 'set_layer_pos_name' to add all positions to the corresponding layers.
            - Calls the 'display_branch' function to build and print the current branch.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_size, int):
            raise InstanceIntegerError("size", argument_size)

        if not isinstance(argument_line, str):
            raise InstanceStringError("line", argument_line)

        if not isinstance(argument_split, str):
            raise InstanceStringError("split", argument_split)

        if not isinstance(argument_end, str):
            raise InstanceStringError("end", argument_end)

        if int(argument_size) < 0 or int(argument_size) > 20:
            raise InvalidSizeError

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    levels = [0]
                    self.elevator(levels, branches[branches_branch][list(branches[branches_branch])[0]])
                    levels.append(0)

                    while True:
                        branch_uid = branches_branch + generate_uid(branches_branch)

                        if branch_uid in list(branches):
                            continue
                        else:
                            uids[branches_branch].remove(branch_uid[-10:])
                            break

                    if not argument_line:
                        argument_line = symbols[branches_branch]["line"]
                    if not argument_split:
                        argument_split = symbols[branches_branch]["split"]
                    if not argument_end:
                        argument_end = symbols[branches_branch]["end"]

                    branches.update({branch_uid: copy.deepcopy(branches[branches_branch])})
                    branches[branch_uid][str("0: " + list(branches[branches_branch])[0])] = \
                        branches[branch_uid].pop(list(branches[branch_uid])[0])
                    output.update({branch_uid: []})
                    uids.update({branch_uid: []})
                    sizes.update({branch_uid: argument_size})
                    symbols.update({branch_uid: {"line": argument_line, "split": argument_split, "end": argument_end}})
                    paint_branch.update({branch_uid: ""})
                    paint_header.update({branch_uid: ""})
                    paint_layer.update({branch_uid: {}})

                    self.set_layer_pos_name(branch_uid, levels, [0],
                                            branches[branch_uid][list(branches[branch_uid])[0]])

                    display_branch(branch_uid, True)

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def set_layer_pos_name(self, branch, levels, trace, value, position_structure=""):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable and the first element of the 'trace' list is
            added with plus '1'.

          Position structure:
            - Then it is checked whether the "level"/"height" of the current value of the loop is equal to or smaller
              than the previous "level"/"height" value of the loop. If the instantaneous value is equal or less, the
              last position and dot in the 'position_structure' variable is removed.
            - Then the current value of the 'position_structure' variable is added with the value of 'count' separated
              by a dot to the 'position_structure' variable.

          - Adds the position to the layer in the current value of 'previous_value'.
          - Checks whether the 'value' variable has a value. If there is a value, then the 'set_layer_pos_name' function
            is called again with the current values of 'value', 'trace' and 'levels' as arguments.
        """
        count = 0
        previous_value = value

        for layer, value in value.copy().items():
            count = count + 1

            trace[0] = trace[0] + 1

            if levels[trace[0]] <= levels[trace[0] - 1]:
                position_structure = position_structure[:-2]
            position_structure = position_structure + "." + str(count)

            previous_value[position_structure[1:] + ": " + layer] = previous_value.pop(layer)

            if value:
                self.set_layer_pos_name(branch, levels, trace, value, position_structure)

    def elevator(self, levels, value, pos=0):
        """
        Value dictionary loop:
          - Loops through each "height" of the value dictionary and adds the value of the 'pos' variable to the 'levels'
            list.
          - Then the 'elevator' function is called again with the current value of the 'value' dictionary.
        """
        for value in value.values():
            levels.append(pos)

            self.elevator(levels, value, pos + 1)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import copy

from br4nch.utility.utility_painter import painter
from br4nch.utility.utility_librarian import paint_branch, paint_header, output, branches, paint_layer, sizes, symbols


class Builder:
    def __init__(self, branch, paint):
        """
        - Gets the arguments and parses them to the 'manager' function.
        """
        self.manager(branch, paint)

    def manager(self, branch, paint):
        """
        - If the 'paint' variable is 'True', then the 'branch_paint', 'header_paint' and 'clear_paint' will be equal to
          the returned values of the painter class.
        - If the 'paint' variable is 'False', then the 'branch_paint', 'header_paint' and 'clear_paint' will be equal to
          empty strings.

        - dictionary, then the corresponding line/output is made for the current value of 'layer'.
        - Calls the 'elevator' function to calculate each level/height of each layer and stores the result in the levels
          list. Then a '0' is added to the 'levels' list so that the 'IndexError' error can be avoided.
        - Calls the 'build_layer' function to create the corresponding output for all layers.
        """
        if paint:
            branch_paint = painter(copy.deepcopy(paint_branch[branch]))
            header_paint = painter(copy.deepcopy(paint_header[branch]))
            clear_paint = painter(["clear"])
        else:
            branch_paint = ""
            header_paint = ""
            clear_paint = ""

        output[branch].append(clear_paint + header_paint + list(branches[branch])[0] + clear_paint + branch_paint +
                              clear_paint)

        levels = [0]
        self.elevator(levels, branches[branch][list(branches[branch])[0]])
        levels.append(0)

        self.build_layer(branch, paint, [branch_paint, clear_paint], [levels, [0], [], [], [0]],
                         branches[branch][list(branches[branch])[0]])

    def elevator(self, levels, value, pos=0):
        """
        Value dictionary loop:
          - Loops through each "height" of the value dictionary and adds the value of the 'pos' variable to the 'levels'
            list.
          - Then the 'elevator' function is called again with the current value of the 'value' dictionary.
        """
        for value in value.values():
            levels.append(pos)

            self.elevator(levels, value, pos + 1)

    def build_layer(self, branch, colored, paint, package, value):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable and the first element of the second element in
            the 'package' list is added with plus '1'.

          - If the 'colored' variable is 'True' and the current value of 'layer' is in the list of 'paint_layer' with
            the branch key, then the 'layer_paint' is equal to the returned values of the painter class.
          - If the 'colored' variable is 'False' or the current value of 'layer' is not in the list of 'paint_layer'
            with the branch key, then the 'layer_paint' is equal to a empty string.

          - Calls the 'build_extender' function to build the corresponding extender for the current value of 'layer'.

          Output:
            - If the value of 'layer' is not equal to the last value of the 'previous_value' dictionary, then the
              corresponding 'split' line/output is made for the current value of 'layer'.
            - If the value of 'layer' is equal to the last value of the 'previous_value' dictionary, then the
              corresponding line/output is made for the current value of 'layer'.

          - Checks whether the 'value' variable has a value. If there is a value, then the 'display_pos' function is
            called again with the current values of 'value', 'trace' and 'levels' as arguments.
        """
        count = 0
        previous_value = value

        for layer, value in value.items():
            count = count + 1

            package[1][0] = package[1][0] + 1

            if colored and layer in paint_layer[branch]:
                layer_paint = painter(copy.deepcopy(paint_layer[branch])[layer])
            else:
                layer_paint = ""

            extend = self.build_extender(branch, layer, previous_value, package[0], package[1], package[2], package[3],
                                         package[4])

            size = ""
            for _ in range(sizes[branch]):
                size = size + extend + symbols[branch].get("line") + "\n"

            if layer != list(previous_value)[-1]:
                output[branch].append(paint[0] + size + extend + symbols[branch].get("split") + " " + paint[1] +
                                      layer_paint + layer[:-15].replace("\n", paint[1] + "\n" + paint[0] + extend +
                                                                        symbols[branch].get("line") + paint[1] + " " *
                                                                        int(len(symbols[branch].get("split"))) +
                                                                        layer_paint) + paint[1])
            else:
                output[branch].append(paint[0] + size + extend + symbols[branch].get("end") + " " + paint[1] +
                                      layer_paint + layer[:-15].replace("\n", paint[1] + "\n" + paint[0] + extend +
                                                                        paint[1] + " " *
                                                                        int(len(symbols[branch].get("end")) + 1) +
                                                                        layer_paint) + paint[1])

            if value:
                self.build_layer(branch, colored, paint, package, value)

    @staticmethod
    def build_extender(branch, layer, previous_value, levels, trace, chain, queue, last):
        """
        Layer height is not equal to '0':
          Layer height is equal to the previous layer height equal:
            Value in 'queue' or first element in 'last' list is equal to the length of all layers in height '0':
              - Adds "x" spaces based on the length of the 'end' symbol and appends it to the 'chain' list and then
                clears the 'queue' list.
              - If the first element in the 'last' list is equal to the length of all layers in height '0', then the
                first element in the 'last' list is equal to '0'.

            - If there is no value in 'queue' or first element in 'last' list is not equal to the length of all layers
              in height '0', then adds a 'line' symbol and "x" spaces based on the length of the 'split' symbol and
              appends it to the 'chain' list.
            - If the layer height is smaller than the previous layer height, the loop will run just as long based on the
              sum of the current layer height minus the next layer's height. With each loop, the last element of the
              'chain' list will be removed.

          - If the value of 'layer' is equal to the last layer in the 'previous_value' list and the height/level of the
            layer is less than the height of the next layer, then the current height of the layer is appended to the
            'queue' list.

        - If the height/level of layer is equal to '0', then the 'chain' list is cleared and the first element in the
          'last' list is added with '1'.

        - Loops through the 'chain' list and composes the 'extend' variable. After the loop, the 'extend' variable is
          returned.
        """
        if levels[trace[0]] != 0:
            if levels[trace[0]] != levels[trace[0] - 1]:
                if queue or last[0] == len(list(branches[branch][list(branches[branch])[0]])):
                    chain.append(" " * int(len(symbols[branch].get("end")) + 1))
                    queue.clear()

                    if last[0] == len(list(branches[branch][list(branches[branch])[0]])):
                        last[0] = 0
                else:
                    chain.append(str(symbols[branch].get("line") + " " * int(len(symbols[branch].get("split")))))

                if levels[trace[0]] < levels[trace[0] - 1]:
                    for _ in range(-1, int(levels[trace[0] - 1] - levels[trace[0]])):
                        chain.pop()

            if layer == list(previous_value)[-1]:
                if levels[trace[0]] < levels[trace[0] + 1]:
                    queue.append(levels[trace[0]])
        else:
            chain.clear()
            last[0] = last[0] + 1

        extend = ""
        for line in chain:
            extend = extend + line
        return extend



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches
from br4nch.utility.utility_printer import printer
from br4nch.utility.utility_positioner import format_position
from br4nch.utility.utility_handler import NotExistingBranchError, InstanceStringError, InstanceBooleanError


def arguments(branch, position, beautify=True):
    """
    - Gets the arguments and parses them to the 'DisplayPosition' class.
    """
    DisplayPosition(branch, position, beautify)


class DisplayPosition:
    def __init__(self, argument_branch, argument_position, argument_beautify):
        """
        - Gets the arguments and parses them to the 'build_position_structure' function.
        """
        self.build_position_structure(argument_branch, argument_position, argument_beautify)

    def build_position_structure(self, argument_branch, argument_position, argument_beautify):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given position argument is not an instance of a list, then the branch argument will be set as a list.

        Errors:
          - If the branch value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            - If the branch is in the 'branches' dictionary, it will loop with all positions in the returned list of the
              'format_position' function. And calls the 'display_position' function for each loop with the built
              position structure and the built 'string_position' variable as arguments.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_position, list):
            argument_position = [argument_position]

        if not isinstance(argument_beautify, bool):
            raise InstanceBooleanError("beautify", argument_beautify)

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    for position in format_position(branches_branch, argument_position.copy()):
                        string_position = ""

                        for character in position:
                            string_position = string_position + "." + character

                        self.display_position(branches_branch, position, string_position[1:], argument_beautify,
                                              branches[branches_branch][list(branches[branches_branch])[0]])

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def display_position(self, branch, position, string_position, argument_beautify, value):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'position':
            If the length of the 'position' list is equal to '1':
              - The 'printer' function is called and a package is supplied with all the values in it that are needed.
                If the 'argument_beautify' variable is false, the given positions are not represented with a branch
                structure.

            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'position' list will be removed and the 'display_position' function will be
              called again with the new value of the 'value' variable as argument.
        """
        count = 0

        for layer, value in value.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    return printer("display_position", [branch, layer[:-15], string_position, argument_beautify])
                else:
                    if value:
                        position.pop(0)
                        return self.display_position(branch, position, string_position, argument_beautify, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import copy

from br4nch.utility.utility_handler import InstanceBooleanError, InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches, uids, paint_layer
from br4nch.utility.utility_positioner import format_position
from br4nch.utility.utility_generator import generate_uid


def arguments(branch, duplicate, position, put="", paint=False, delete=False):
    """
    - Gets the arguments and parses them to the 'DuplicateLayer' class.
    """
    DuplicateLayer(branch, duplicate, position, put, paint, delete)


class DuplicateLayer:
    def __init__(self, argument_branch, argument_duplicate, argument_position, argument_put, argument_paint,
                 argument_delete):
        """
        - Gets the arguments and parses them to the 'duplicate_layer' function.
        """
        self.duplicate_layer(argument_branch, argument_duplicate, argument_position, argument_put, argument_paint,
                             argument_delete)

    def duplicate_layer(self, argument_branch, argument_duplicate, argument_position, argument_put, argument_paint,
                        argument_delete):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given duplicate argument is not an instance of a list, then the duplicate argument will be set as a
            list.
          - If the given position argument is not an instance of a list, then the position argument will be set as a
            list.
          - If the given put argument is not an instance of a list, then the put argument will be set as a list.

        Errors:
          - If the paint value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
          - If the delete value is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.
          - If there a '*' in the 'argument_put' list, Then it appends all existing branches to the 'argument_put' list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            Argument duplicate list loop:
              Argument position list loop:
                - Calls the function 'task_manager' to perform the necessary tasks for the variable 'argument_copy'.

                If package:
                  - Adds the second value in the returned package to the list 'queue_delete'.

                  - If the 'put' argument is not used. The 'argument_put' list will be changed to the current branch
                    value in a list.

                  Argument put list loop:
                    Errors:
                      - If the branch value is not an instance of a string, then it raises an 'InstanceStringError'
                        error.
                      - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError'
                        error.

                    Branches list loop:
                      - Adds the first value in the returned package and calls the function 'task_manager' to perform
                        the necessary tasks for the variable 'argument_position' and adds the returned dictionary in a
                        list to the list 'queue_add'.

            - Loops through the list 'queue_delete' and deletes all given layers from the dictionary 'branches'.
            - Loops through the list 'queue_add' and adds all given layers from the dictionary 'branches'.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_duplicate, list):
            argument_duplicate = [argument_duplicate]

        if not isinstance(argument_position, list):
            argument_position = [argument_position]

        if not isinstance(argument_put, list):
            argument_put = [argument_put]

        if not isinstance(argument_delete, bool):
            raise InstanceBooleanError("paint", argument_paint)

        if not isinstance(argument_delete, bool):
            raise InstanceBooleanError("delete", argument_delete)

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        if "*" in argument_put:
            argument_put.clear()
            for branches_branch in list(branches):
                argument_put.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    queue_delete = []
                    queue_add = []

                    for loop_copy in format_position(branches_branch, argument_duplicate.copy()):
                        for loop_position in format_position(branches_branch, argument_position.copy()):
                            package = self.task_manager(branches_branch, loop_copy, [], argument_paint,
                                                        branches[branches_branch][list(branches[branches_branch])[0]])

                            if package:
                                queue_delete.append(package[1])

                                if len(argument_put) == 1 and not argument_put[0]:
                                    argument_put = [branches_branch]

                                for put_branch in argument_put:
                                    put_error = 0

                                    if not isinstance(put_branch, str):
                                        raise InstanceStringError("put", put_branch)

                                    for branches_branch_two in list(branches):
                                        if put_branch.lower() == branches_branch_two.lower():
                                            put_error = put_error + 1

                                            queue_add.append([package[0], self.task_manager(branches_branch_two, [],
                                                                                            loop_position,
                                                                                            argument_paint,
                                                                                            branches
                                                                                            [branches_branch_two]
                                                                                            [list(branches
                                                                                                  [branches_branch_two])
                                                                                            [0]])])

                                    if put_error == 0:
                                        if put_branch:
                                            raise NotExistingBranchError(put_branch)

                    if argument_delete:
                        for delete_value in queue_delete:
                            if delete_value:
                                for layer, value in delete_value.items():
                                    uids[branches_branch].remove(str(layer[-10:]))
                                    paint_layer[branches_branch].pop(layer)

                                    self.delete_layer_additions(branches_branch, value[layer])

                                    del value[layer]

                    for add_value in queue_add:
                        if add_value and add_value[0] and add_value[1]:
                            self.change_layer_uid(list(add_value[1])[0], argument_paint, add_value[0])
                            add_value[1][list(add_value[1])[0]].update(add_value[0])

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def task_manager(self, branch, duplicate, position, argument_paint, value):
        """
        - If the first 'position' element variable is equal to '0', then it returns the whole branch from the 'branches'
          dictionary.

        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'duplicate':
            If the length of the 'duplicate' list is equal to '1':
              - Adds the current layer and aa deepcopy of the 'value' variable to a dictionary and adds the current
                layer and previous value to a dictionary and returns these two values as a package.

            - If the length of the 'duplicate' list is not equal to '1' and there is a value of the 'value'
              variable, then the first value from the 'argument_move' list will be removed and the 'task_manager'
              function will be called again with the new value of the 'value' variable as argument.

          Count variable equal to the first value of 'position':
            If the length of the 'position' list is equal to '1':
              - Returns the current value of the 'value' variable.

            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'position' list will be removed and the 'task_manager' function will be
              called again with the new value of the 'value' variable as argument.
        """
        count = 0
        previous_value = value

        if position and position[0] == "0":
            return {branch: branches[branch][list(branches[branch])[0]]}

        for layer, value in value.items():
            count = count + 1

            if duplicate and count == int(duplicate[0]):
                if len(duplicate) == 1:
                    return [{layer: copy.deepcopy(value)}, {layer: previous_value}]
                else:
                    if value:
                        duplicate.pop(0)
                        return self.task_manager(branch, duplicate, position, argument_paint, value)

            if position and count == int(position[0]):
                if len(position) == 1:
                    return {branch: value}
                else:
                    if value:
                        position.pop(0)
                        return self.task_manager(branch, duplicate, position, argument_paint, value)

    def delete_layer_additions(self, branch, value):
        """
        Value dictionary loop:
          - Checks if the UID of the current value of 'layer' exists in the 'uids' list. If the UID is in the 'uids'
            list, then it will be removed from the list.
          - Checks if the 'layer' value exists in the 'paint_layer' branch list. If the layer is in the 'paint_layer'
            branch list, then the value of the current layer key in the 'paint_layer' dictionary will be removed.
          - If there is a value of the 'value' variable, the 'delete_layer_additions' function will be called again with
            the new value of the 'value' variable as argument.
        """
        for layer, value in value.items():
            if layer[-10:] in uids[branch]:
                uids[branch].remove(str(layer[-10:]))

            if layer in paint_layer[branch]:
                paint_layer[branch].pop(layer)

            if value:
                self.delete_layer_additions(branch, value)

    def change_layer_uid(self, branch, argument_paint, value):
        """
        Value dictionary loop:
          - Generates a new UID for the copied 'layer' variable. Then the old layer is deleted and replaced with the new
            generated layer with the new UID.

          - Checks if the current value of 'layer' exists in the branch's 'paint_layer' list. Then it is checked whether
            the variable 'argument_paint' is true. If the value exists in the list and the 'argument_paint' is true,
            then the newly generated layer with UID is added to the list with the value of the value of the variable
            'layer'.
          - If the current value of 'layer' does not exist in the branch's 'paint_layer' list. Then the newly generated
            layer with UID is added to the list with the value of an empty list.

          - If there is a value of the 'value' variable, the 'change_layer_uid' function will be called again with the
            new value of the 'value' variable as argument.
        """
        previous_value = value

        for layer, value in value.copy().items():
            new_layer = layer[:-15] + generate_uid(branch)

            previous_value[new_layer] = previous_value.pop(layer)

            if layer in paint_layer[branch]:
                if argument_paint:
                    paint_layer[branch].update({new_layer: paint_layer[branch][layer]})
            else:
                paint_layer[branch].update({new_layer: []})

            if value:
                self.change_layer_uid(branch, argument_paint, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import MaximumPaintSlotsError


def painter(paint):
    """
    Errors:
      - If the length of the 'paint' list is bigger than '4', then it raises a 'MaximumPaintSlotsError' error.

    - Then is calculated how many times the loop should be run and adding element with an empty string as value to the
      list 'paint' until the length of the list is equal to '4'.

    Length 'paint' list loop:
      - The current value of position in the 'paint' list will be set as the requested paint.

    - Then all paint is returned.
    """
    if len(paint) > 4:
        raise MaximumPaintSlotsError

    for _ in range(4 - len(paint)):
        paint.append("")

    for position in range(len(paint)):
        if paint[position]:
            paint[position] = ["\u001b[31m", "\u001b[33m", "\u001b[32m", "\u001b[36m", "\u001b[34m", "\u001b[35m",
                               "\u001b[30m", "\u001b[37m", "\u001b[1m", "\u001b[4m", "\u001b[7m", "\u001b[0m"][
                ["red", "yellow", "green", "cyan", "blue", "magenta", "black", "white", "bold", "underline",
                 "reversing", "clear"].index(paint[position].lower())]

    return paint[0] + paint[1] + paint[2] + paint[3]



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches, paint_layer
from br4nch.utility.utility_handler import InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_positioner import format_position
from br4nch.utility.utility_generator import generate_uid


def arguments(branch, layer, position):
    """
    - Gets the arguments and parses them to the 'CreateLayer' class.
    """
    CreateLayer(branch, layer, position)


class CreateLayer:
    def __init__(self, argument_branch, argument_layer, argument_position):
        """
        - Gets the arguments and parses them to the 'build_position_structure' function.
        """
        self.build_position_structure(argument_branch, argument_layer, argument_position)

    def build_position_structure(self, argument_branch, argument_layer, argument_position):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given layer argument is not an instance of a list, then the layer argument will be set as a list.
          - If the given position argument is not an instance of a list, then the position argument will be set as a
            list.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            Errors:
              - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
              - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

            - If the branch is in the 'branches' dictionary, then it runs a loop with all positions in the returned list
              from the 'format_position' function. And calls the 'add_layer' function with the whole branch dictionary
              as value for every looped position.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_layer, list):
            argument_layer = [argument_layer]

        if not isinstance(argument_position, list):
            argument_position = [argument_position]

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    for position in format_position(branches_branch, argument_position):
                        self.create_layer(branches_branch, argument_layer, position,
                                          branches[branches_branch][list(branches[branches_branch])[0]])

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def create_layer(self, branch, argument_layer, position, value):
        """
        Position variable equal to zero:
          Errors:
            - If the 'loop_layer' value is not an instance of a string, then it raises an 'InstanceStringError' error.

          - If the last character in 'loop_layer' is equal to a newline/'\n', then it removes it from 'loop_layer'.

          - If the first value in the 'position' is equal to a '0' then the 'argument_layer' variable is looped and in
            each loop it will create the current layer with a generated UID to the value to the 'paint_layer' list and
            to the 'value' dictionary from the 'branches' dictionary.

        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'position':
            If the length of the 'position' list is equal to '1':
              Errors:
                - If the layer value is not an instance of a string, then it raises an 'InstanceStringError' error.

              - If the last character in 'loop_layer' is equal to a newline/'\n', then it removes it from 'loop_layer'.

              - Then it will create the current layer with a generated UID to the value to the 'paint_layer' list and to
                the 'value' dictionary from the 'branches' dictionary.

            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'position' list will be removed and the 'add_layer' function will be called
              again with the new value of the 'value' variable as argument.
        """
        if position[0] == "0":
            for loop_layer in argument_layer:
                if not isinstance(loop_layer, str):
                    raise InstanceStringError("layer", loop_layer)

                while True:
                    if loop_layer and loop_layer[-1] == "\n":
                        loop_layer = loop_layer[:-1]
                    else:
                        break

                uid_layer = loop_layer + generate_uid(branch)

                paint_layer[branch].update({uid_layer: []})
                branches[branch][list(branches[branch])[0]].update({uid_layer: {}})
            return

        count = 0

        for value in value.values():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    for loop_layer in argument_layer:
                        if not isinstance(loop_layer, str):
                            raise InstanceStringError("layer", loop_layer)

                        while True:
                            if loop_layer and loop_layer[-1] == "\n":
                                loop_layer = loop_layer[:-1]
                            else:
                                break

                        uid_layer = loop_layer + generate_uid(branch)

                        paint_layer[branch].update({uid_layer: []})
                        value.update({uid_layer: {}})
                    return
                else:
                    if value:
                        position.pop(0)
                        return self.create_layer(branch, argument_layer, position, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, InvalidPositionError
from br4nch.utility.utility_librarian import branches


def format_position(branch, position_package):
    """
    position_package range loop:
      - If the position value is an instance of a list.
        position_package[number] range loop:
          - If the position value is not an instance of a string, then it raises an 'InstanceStringError' error.
          - If the character variable in the position loop is not equal to a number and/or operator, the error
            'InvalidPositionError' is displayed.

      - If the position value is not an instance of a list.
          - If the position_package[number] value is not an instance of a string, then it raises an
            'InstanceStringError' error.
          - If the character variable in the position_package[number] loop is not equal to a number and/or operator, the
            error 'InvalidPositionError' is displayed.

      - If there is a '.' in the 'position_package[number]', then the 'position_package[number]' variable is equal to
        the split 'position_package[number]' list.
      - If there is no '.' in the 'position_package[number]', if the given pos argument is not an instance of a list,
        then the pos argument is set as a list.

    position_package range loop:
      position_package[number] range loop:
        - If there is a '/' in the 'position_package[number][position]' variable, The
          'position_package[number][position]' will be split at the '/' character. If there is a element in the list
          without a value, the 'InvalidPositionError' error will be called. if all the elements in the list has a value,
          a loop is made with the value of the length of the split values. Then in each loop it makes a copy of the
          current value of the 'position_package[number][position]' variable and adds the value to the
          'position_package' list and changes the position to the copied value with the position value of the split
          'position_package[number][position]' value. Then the value of 'number' is removed in the 'position_package'
          list.
        - Calls the 'format_position' function again.

    position_package range loop:
      position_package[number] range loop:
        - If there is a '*' in the 'position_package[number][position]' variable, if there are no existing layers in the
          current branch, then the '*' position will be changed to a '0'. If there are existing layer in the current
          branch, then a loop is made with the value of the length of the returned total layers value in the
          'calculate_operator' function. Then in each loop it makes a copy of the current value of the
          'position_package[number][position]' variable and adds the value to the 'position_package' list. Then it
          changes the position to the copied value with the current number of 'count + 1' value. Then the value of
          'number' is removed in the 'position_package' list.

        - If there is a '>' in the variable 'position_package[number][position]', then the
          'position_package[number][position]' is split and stored in the variable 'including_positions'. Then the
          length of 'including_positions' is also stored in a variable called 'total_including_positions'.
        - A loop is made with the lowest value of 'include_positions' and the highest value of 'include_positions' plus
          one. In each loop, the value 'count' is added to the list 'including_positions'.
        - Then a loop is made of the length of the 'including_positions' variable. In each loop, the first value of the
          'including_positions' list will be removed.
        - A loop is made with the value of the length of the total length of the 'including_positions' variable. Then in
          each loop it makes a copy of the current value of the 'position_package[number][position]' variable and adds
          the value to the 'position_package' list. Then it changes the position to the copied value with the current
          number of 'count' in the 'including_positions' list. Then the value of 'number' is removed in the
          'position_package' list.

        - If there is a '<' in the 'position_package[number][position]' variable, then a loop is made with the value of
          the length of the returned total layers value in the 'calculate_operator' function. Then in each loop it
          appends the string 'count + 1' to the 'excluding_positions' list.
        - A loop is made with the lowest value of the split 'position_package[number][position]' and the highest value
          of the split 'position_package[number][position]' plus one. In each loop, the string value of 'count' will be
          removed from the 'excluding_positions' list.
        - A loop is made with the value of the length of the total length of the 'excluding_positions' variable. Then in
          each loop it makes a copy of the current value of the 'position_package[number][position]' variable and adds
          the value to the 'position_package' list. Then it changes the position to the copied value with the current
          number of 'count' in the 'excluding_positions' list. Then the value of 'number' is removed in the
          'position_package' list.

    - Calls the 'format_position' function to manage the rest of the operators if there is a '/', '*', '>' or '<' in the
      'position_package[number][position]' variable.

    - Returns the 'position_package' variable.
    """
    for number in range(len(position_package)):
        if isinstance(position_package[number], list):
            for position in position_package[number]:
                if not isinstance(position, str):
                    raise InstanceStringError("position", position)
                else:
                    for character in position:
                        if character not in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ".", "/", "*", ">", "<"]:
                            print(position_package[number])
                            raise InvalidPositionError("position", position_package[number])
        else:
            if not isinstance(position_package[number], str):
                raise InstanceStringError("pos", position_package[number])
            else:
                for character in position_package[number]:
                    if character not in ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ".", "/", "*", ">", "<"]:
                        raise InvalidPositionError("position", position_package[number])

        if "." in position_package[number]:
            position_package[number] = position_package[number].split(".")
        else:
            if not isinstance(position_package[number], list):
                position_package[number] = [position_package[number]]

    for number in range(len(position_package)):
        for position in range(len(position_package[number])):
            if "/" in position_package[number][position]:
                for multiple_position in position_package[number][position].split("/"):
                    if not multiple_position:
                        raise InvalidPositionError("position", position_package[number][position])

                for count in range(len(position_package[number][position].split("/"))):
                    position_package.append(position_package[number].copy())
                    position_package[-1][position] = position_package[number][position].split("/")[count]
                position_package.pop(number)

                format_position(branch, position_package)

    for number in range(len(position_package)):
        for position in range(len(position_package[number])):
            if "*" in position_package[number][position]:
                if not branches[branch][list(branches[branch])[0]]:
                    position_package[number].pop(position)
                    position_package[number].append("0")
                else:
                    for count in range(calculate_operator(position_package[number].copy(),
                                                          branches[branch][list(branches[branch])[0]])):
                        position_package.append(position_package[number].copy())
                        position_package[-1][position] = str(count + 1)
                    position_package.pop(number)

            if ">" in position_package[number][position]:
                if ">" in position_package[number][position]:
                    for including_position in position_package[number][position].split(">"):
                        if not including_position:
                            raise InvalidPositionError("position", position_package[number][position])

                including_positions = position_package[number][position].split(">")
                total_including_positions = len(including_positions)

                for count in range(int(min(including_positions)), int(max(including_positions)) + 1):
                    including_positions.append(str(count))

                for _ in range(total_including_positions):
                    including_positions.pop(0)

                for count in range(len(including_positions)):
                    position_package.append(position_package[number].copy())
                    position_package[-1][position] = including_positions[count]
                position_package.pop(number)

            if "<" in position_package[number][position]:
                if "<" in position_package[number][position]:
                    for excluding_position in position_package[number][position].split("<"):
                        if not excluding_position:
                            raise InvalidPositionError("position", position_package[number][position])

                excluding_positions = []

                for count in range(calculate_operator(position_package[number].copy(),
                                                      branches[branch][list(branches[branch])[0]])):
                    excluding_positions.append(str(count + 1))

                for count in range(int(min(position_package[number][position].split("<"))),
                                   int(max(position_package[number][position].split("<"))) + 1):
                    excluding_positions.remove(str(count))

                for count in range(len(excluding_positions)):
                    position_package.append(position_package[number].copy())
                    position_package[-1][position] = excluding_positions[count]
                position_package.pop(number)

            if "/" in position_package[number][position] or "*" in position_package[number][position]\
                    or ">" in position_package[number][position] or "<" in position_package[number][position]:
                format_position(branch, position_package)

    return position_package


def calculate_operator(position, value):
    """
    Value dictionary loop:
      - For each value of the 'value' variable the 'count' variable is added with plus '1'.

      - If the '*' or '<' is in the first element in the 'position' list, or if the 'count' variable is equal to the
        first element in the 'position' list as an integer, then all layers are returned.

      - If there is a value of the 'value' variable and the 'count' variable is equal to the first element in the
        'position' list as an integer, then the first value is removed from the 'position' list and the
        'calculate_operator' function is called again with the new values of the variable 'value' as argument.
    """
    count = 0
    previous_value = value

    for layer, value in value.items():
        count = count + 1

        if "*" in position[0] or "<" in position[0] or count == int(position[0]):
            return len(previous_value)

        if value and count == int(position[0]):
            position.pop(0)
            return calculate_operator(position, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches, uids, paint_layer
from br4nch.utility.utility_positioner import format_position
from br4nch.utility.utility_generator import generate_uid


def arguments(branch, position, name):
    """
    - Gets the arguments and parses them to the 'ReplaceLayer' class.
    """
    ReplaceLayer(branch, position, name)


class ReplaceLayer:
    def __init__(self, argument_branch, argument_position, argument_name):
        """
        - Gets the arguments and parses them to the 'replace_layer' function.
        """
        self.replace_layer(argument_branch, argument_position, argument_name)

    def replace_layer(self, argument_branch, argument_position, argument_name):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given position argument is not an instance of a list, then the position argument will be set as a
            list.

        Errors:
          - If the name argument is not an instance of a string, then it raises an 'InstanceStringError' error.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Branches list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Argument copy list loop:
            - Calls the function 'task_manager' to perform the necessary tasks for the variable 'argument_copy'
            - If the last character in 'argument_name' is equal to a newline/'\n', then it removes it from
              'argument_name'.
            - Loops through the items from the returned dictionary and replaces all given layers from the mandatory
              dictionaries.
            - To keep the order the same, a loop will be made that loops through the 'value' dictionary and
              'paint_layer' list with the given slice values, removing all values and adding them again.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_position, list):
            argument_position = [argument_position]

        if not isinstance(argument_name, str):
            raise InstanceStringError("name", argument_name)

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    for position in format_position(branches_branch, argument_position.copy()):
                        returned_value = self.get_layers(branches_branch, position,
                                                         branches[branches_branch][list(branches[branches_branch])[0]])
                        if returned_value:
                            for layer, value in returned_value.items():
                                while True:
                                    if argument_name and argument_name[-1] == "\n":
                                        argument_name = argument_name[:-1]
                                    else:
                                        break

                                new_layer = argument_name + generate_uid(branches_branch)
                                uids[branches_branch].remove(layer[-10:])
                                paint_layer[branches_branch][new_layer] = paint_layer[branches_branch].pop(layer)

                                index = list(value).index(layer)
                                value[new_layer] = value.pop(layer)

                                for number in list(value)[index:-1]:
                                    value[number] = value.pop(number)
                                    paint_layer[branches_branch][number] = paint_layer[branches_branch].pop(number)

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def get_layers(self, branch, position, value):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'position':
            If the length of the 'position' list is equal to '1':
              - Returns the current layer and previous value in an dictionary.

            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'argument_move' list will be removed and the 'get_layers' function will be
              called again with the new value of the 'value' variable as argument.
        """
        count = 0
        previous_value = value

        for layer, value in value.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    return {layer: previous_value}
                else:
                    if value:
                        position.pop(0)
                        return self.get_layers(branch, position, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches


def arguments(branch, name):
    """
    - Gets the arguments and parses them to the 'replace_header' function.
    """
    replace_header(branch, name)


def replace_header(argument_branch, argument_name):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the name argument is not an instance of a string, then it raises an 'InstanceStringError' error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Branches list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      - Deletes the given header and creates a new one with the given name and the copied values from the given header.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_name, str):
        raise InstanceStringError("name", argument_name)

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                branches[branches_branch][argument_name] = \
                    branches[branches_branch].pop(list(branches[branches_branch])[0])

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, InstanceBooleanError, InvalidPositionError, \
    NotExistingBranchError
from br4nch.utility.utility_librarian import branches, paint_layer
from br4nch.utility.utility_positioner import format_position
from br4nch.utility.utility_generator import generate_uid


def arguments(branch, move, position, put="", paint=False):
    """
    - Gets the arguments and parses them to the 'MoveLayer' class.
    """
    MoveLayer(branch, move, position, put, paint)


class MoveLayer:
    def __init__(self, argument_branch, argument_move, argument_position, argument_put, argument_paint):
        """
        - Gets the arguments and parses them to the 'move_layer' function.
        """
        self.move_layer(argument_branch, argument_move, argument_position, argument_put, argument_paint)

    def move_layer(self, argument_branch, argument_move, argument_position, argument_put, argument_paint):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given move argument is not an instance of a list, then the move argument will be set as a list.

        Errors:
          - If the position argument is not an instance of a string, then it raises an 'InstanceStringError' error.
          - If the put argument is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
          - If the paint argument is not an instance of a boolean, then it raises an 'InstanceBooleanError' error.
          - If the position argument is not equal to a number and/or operator, then it raises an 'InvalidPositionError'
            error.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            Argument copy list loop:
              - Calls the function 'task_manager' to perform the necessary tasks for the variable 'argument_move'.

                If package:
                  - Adds the second value in the returned package to the list 'queue_delete'.
                  - The branches_branch value will be used as branch to move the chosen position(s) to.

                  If argument_put:
                    Branches list loop:
                      Errors:
                        - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError'
                          error.

                      - The argument_put value will be used as branch to move the chosen position(s) to.

              - Adds the first value in the returned package and calls the function 'task_manager' to perform the
                necessary tasks for the variable 'argument_position' and adds the returned dictionary in a list to the
                list 'queue_add'.

          - Loops through the list 'queue_delete' and deletes all given layers from the dictionary 'branches'.
          - Loops through the list 'queue_add' and adds all given layers from the dictionary 'branches'.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_move, list):
            argument_move = [argument_move]

        if not isinstance(argument_put, str):
            raise InstanceStringError("put", argument_put)

        if not isinstance(argument_position, str):
            raise InstanceStringError("position", argument_position)

        if not isinstance(argument_paint, bool):
            raise InstanceBooleanError("paint", argument_paint)

        for position in argument_position.split("."):
            if not position.isnumeric():
                raise InvalidPositionError("position", argument_position)

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    queue_delete = []
                    queue_add = []

                    for position in format_position(branches_branch, argument_move.copy()):
                        package = self.task_manager(branches_branch, position, [], {},
                                                    branches[branches_branch][list(branches[branches_branch])[0]])

                        if package:
                            queue_delete.append(package[1])

                            move_branch = branches_branch

                            if argument_put:
                                put_error = 0

                                for branches_branch_two in list(branches):
                                    if argument_put.lower() == branches_branch_two.lower():
                                        put_error = put_error + 1

                                        move_branch = argument_put

                                if put_error == 0:
                                    if argument_put:
                                        raise NotExistingBranchError(argument_put)

                            queue_add.append([package[0], self.task_manager(move_branch, [],
                                                                            argument_position.split("."), package[0],
                                                                            branches[move_branch]
                                                                            [list(branches[move_branch])[0]])])

                    for delete_value in queue_delete:
                        if delete_value:
                            for layer, value in delete_value.items():
                                del value[layer]

                    for add_value in queue_add:
                        if add_value and add_value[0] and add_value[1]:
                            self.change_layer_uid(list(add_value[1])[0], argument_paint, add_value[0])
                            add_value[1][list(add_value[1])[0]].update(add_value[0])

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def task_manager(self, branch, move, position, copied_layer, value):
        """
        - If the first 'position' element variable is equal to '0', then it returns the whole branch from the 'branches'
          dictionary.

        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'move':
            If the length of the 'move' list is equal to '1':
              - Adds the current layer and value to a dictionary and adds the current layer and previous value to a
                dictionary and returns these two values as a package.

            - If the length of the 'move' list is not equal to '1' and there is a value of the 'value'
              variable, then the first value from the 'move' list will be removed and the 'task_manager'
              function will be called again with the new value of the 'value' variable as argument.

          Count variable equal to the first value of 'position':
            If the length of the 'position' list is equal to '1':
              - Returns the current value of the 'value' variable.

            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'position' list will be removed and the 'task_manager' function will be
              called again with the new value of the 'value' variable as argument.
        """
        count = 0
        previous_value = value

        if position and position[0] == "0":
            return {branch: branches[branch][list(branches[branch])[0]]}

        for layer, value in value.items():
            count = count + 1

            if move and count == int(move[0]):
                if len(move) == 1:
                    return [{layer: value}, {layer: previous_value}]
                else:
                    if value:
                        move.pop(0)
                        return self.task_manager(branch, move, position, copied_layer, value)

            if position and count == int(position[0]):
                if len(position) == 1:
                    return {branch: value}
                else:
                    if value:
                        position.pop(0)
                        return self.task_manager(branch, move, position, copied_layer, value)

    def change_layer_uid(self, branch, argument_paint, value):
        """
        Value dictionary loop:
          - Generates a new UID for the copied 'layer' variable. Then the old layer is deleted and replaced with the new
            generated layer with the new UID.

          - Checks if the current value of 'layer' exists in the branch's 'paint_layer' list. Then it is checked whether
            the variable 'argument_paint' is true. If the value exists in the list and the 'argument_paint' is true,
            then the newly generated layer with UID is added to the list with the value of the value of the variable
            'layer'.
          - If the current value of 'layer' does not exist in the branch's 'paint_layer' list. Then the newly generated
            layer with UID is added to the list with the value of an empty string.

          - If there is a value of the 'value' variable, the 'change_layer_uid' function will be called again with the
            new value of the 'value' variable as argument.
        """
        previous_value = value

        for layer, value in value.copy().items():
            new_layer = layer[:-15] + generate_uid(branch)

            previous_value[new_layer] = previous_value.pop(layer)

            if layer in paint_layer[branch]:
                if argument_paint:
                    paint_layer[branch].update({new_layer: paint_layer[branch][layer]})
            else:
                paint_layer[branch].update({new_layer: []})

            if value:
                self.change_layer_uid(branch, argument_paint, value)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import br4nch.utility.utility_builder
import br4nch.utility.utility_librarian
import br4nch.utility.utility_generator
import br4nch.utility.utility_positioner
import br4nch.utility.utility_painter
import br4nch.utility.utility_handler



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, InvalidBranchNameError, DuplicateBranchError
from br4nch.utility.utility_librarian import branches, output, uids, sizes, symbols, paint_branch, paint_header, \
    paint_layer


def arguments(branch, header):
    """
    - Gets the arguments and parses them to the 'create_branch' function.
    """
    create_branch(branch, header)


def create_branch(argument_branch, argument_header):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the header value is not an instance of a string, then it raises an 'InstanceStringError' error.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch value contains a character that is not a letter or number, then it raises an
          'InvalidBranchError' error.

      Branches list loop:
        Errors:
          - If the branch is already in the 'branches' dictionary, then it raises a 'DuplicateBranchError' error.

      - Then it will create the current branch key in all the mandatory dictionaries with the needed value.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_header, str):
        raise InstanceStringError("header", argument_header)

    for branch in argument_branch:
        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        if not branch.isalnum():
            raise InvalidBranchNameError(branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                raise DuplicateBranchError(branch)

        branches.update({branch: {argument_header: {}}})
        output.update({branch: []})
        uids.update({branch: []})
        sizes.update({branch: 0})
        symbols.update({branch: {"line": "┃", "split": "┣━", "end": "┗━"}})
        paint_branch.update({branch: []})
        paint_header.update({branch: []})
        paint_layer.update({branch: {}})



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

import os

from br4nch.utility.utility_handler import InstanceStringError, NotExistingDirectoryError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches, output
from br4nch.utility.utility_builder import Builder


def arguments(branch, directory):
    """
    - Gets the arguments and parses them to the 'export_text' function.
    """
    export_text(branch, directory)


def export_text(argument_branch, argument_directory):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the directory value is not an instance of a string, then it raises an 'InstanceStringError' error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If and the branch value is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      Branches list loop:
        Errors:
          - If the given directory argument does not exist, it will throw a 'NotExistingDirectoryError' error.

        If the given directory exists:
          - Creates a new file with a custom name of the value of branch
          - Calls the 'Builder' class to create the branch's 'output' variable.
          - Loops through the output of the branch and appends each element in the last to the file with a new line.
          - Clears the output list from the value of branch.
    """

    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_directory, str):
        raise InstanceStringError("directory", argument_directory)

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                if os.path.isdir(argument_directory):
                    with open(argument_directory + "/br4nch-" + branches_branch + ".txt", 'w', encoding='utf-8')\
                            as file:
                        Builder(branches_branch, False)
                        for line in output[branches_branch]:
                            file.write(line + "\n")
                        output[branches_branch].clear()
                else:
                    raise NotExistingDirectoryError(argument_directory)

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches, paint_branch
from br4nch.utility.utility_handler import InstanceStringError, NotExistingPaintError, NotExistingBranchError


def arguments(branch, paint):
    """
    - Gets the arguments and parses them to the 'set_paint_branch' function.
    """
    set_paint_branch(branch, paint)


def set_paint_branch(argument_branch, argument_paint):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
      - If the given paint argument is not an instance of a list, then the paint argument will be set as a list.

    Errors:
      - If the paint value is not an instance of a string, then it raises an 'InstanceStringError' error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch value is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      Branches list loop:
        Errors:
          - If the paint element argument is not an instance of a string, then it raises an 'InstanceStringError' error.

        - If the branch is in the 'branches' dictionary, then it check if the 'paint' exists in the given list. If it
          does not exists, it raises a 'NotExistingPaintError' error. If it does exists in the list, it will add the
          'argument_paint' list to the 'paint_branch' list.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_paint, list):
        argument_paint = [argument_paint]

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                for paint in argument_paint:
                    if not isinstance(paint, str):
                        raise InstanceStringError("paint", paint)

                    if paint.lower() not in ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white",
                                             "bold", "underline", "reversing"]:
                        raise NotExistingPaintError(paint)

                paint_branch.update({branches_branch: argument_paint})

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, InvalidBranchNameError, DuplicateBranchError, \
    NotExistingBranchError
from br4nch.utility.utility_librarian import branches, output, uids, sizes, symbols, paint_branch, paint_header, \
    paint_layer


def arguments(branch, name):
    """
    - Gets the arguments and parses them to the 'replace_branch' function.
    """
    replace_branch(branch, name)


def replace_branch(argument_branch, argument_name):
    """
    Errors:
      - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If the name value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If the name value contains a character that is not a letter or number, then it raises an
        'InvalidBranchError' error.

    Branches list loop:
      Errors:
        - If the branch is already in the 'branches' dictionary, then it raises a 'DuplicateBranchError' error.

    Branches list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      - The index of the given branch in the 'branches' list is stored in the variable 'index'.
      - Deletes the given branch and creates a new one with the given name and the copied values from the given branch.
      - To keep the order the same, a loop will be made that loops through the branches with the given slice values,
        removing all values and adding them again.
    """
    if not isinstance(argument_branch, str):
        raise InstanceStringError("branch", argument_branch)

    if not isinstance(argument_name, str):
        raise InstanceStringError("name", argument_name)

    if not argument_name.isalnum():
        raise InvalidBranchNameError(argument_name)

    for branches_branch in list(branches):
        if argument_name.lower() == branches_branch.lower():
            raise DuplicateBranchError(argument_name)

    error = 0
    for branches_branch in list(branches):
        if argument_branch.lower() == branches_branch.lower():
            error = error + 1

            index = list(branches).index(branches_branch)

            branches[argument_name] = branches.pop(branches_branch)
            output[argument_name] = output.pop(branches_branch)
            sizes[argument_name] = sizes.pop(branches_branch)
            symbols[argument_name] = symbols.pop(branches_branch)
            paint_branch[argument_name] = paint_branch.pop(branches_branch)
            paint_header[argument_name] = paint_header.pop(branches_branch)
            paint_layer[argument_name] = paint_layer.pop(branches_branch)
            uids[argument_name] = uids.pop(branches_branch)

            for position in list(branches)[index:-1]:
                branches[position] = branches.pop(position)
                output[position] = output.pop(position)
                sizes[position] = sizes.pop(position)
                symbols[position] = symbols.pop(position)
                paint_branch[position] = paint_branch.pop(position)
                paint_header[position] = paint_header.pop(position)
                paint_layer[position] = paint_layer.pop(position)
                uids[position] = uids.pop(position)

    if error == 0:
        if argument_branch:
            raise NotExistingBranchError(argument_branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_handler import InstanceStringError, RequiredSymbolChangeError, NotExistingBranchError
from br4nch.utility.utility_librarian import branches, symbols


def arguments(branch, line="", split="", end=""):
    """
    - Gets the arguments and parses them to the 'set_symbol' function.
    """
    set_symbol(branch, line, split, end)


def set_symbol(argument_branch, argument_line, argument_split, argument_end):
    """
    Lists:
      - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.

    Errors:
      - If the line value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If the split value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If the end value is not an instance of a string, then it raises an 'InstanceStringError' error.
      - If there is no value in the 'line', 'split' and 'end' arguments, then it raises an 'RequiredSymbolChangeError'
        error.

    Operators:
      - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
        list.

    Argument branch list loop:
      Errors:
        - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
        - If the branch value is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

      Branches list loop:
        - If there is no value in any of the 'line', 'split' or 'end' arguments, the value that is currently active of
          the missing arguments is used.
        - If the branch is in the 'branches' dictionary, then all values of the symbol arguments are added to the
          required directories.
    """
    if not isinstance(argument_branch, list):
        argument_branch = [argument_branch]

    if not isinstance(argument_line, str):
        raise InstanceStringError("line", argument_line)

    if not isinstance(argument_split, str):
        raise InstanceStringError("split", argument_split)

    if not isinstance(argument_end, str):
        raise InstanceStringError("end", argument_end)

    if not argument_line and not argument_split and not argument_end:
        raise RequiredSymbolChangeError

    if "*" in argument_branch:
        argument_branch.clear()
        for branches_branch in list(branches):
            argument_branch.append(branches_branch)

    for branch in argument_branch:
        error = 0

        if not isinstance(branch, str):
            raise InstanceStringError("branch", branch)

        for branches_branch in list(branches):
            if branch.lower() == branches_branch.lower():
                error = error + 1

                if not argument_line:
                    argument_line = symbols[branches_branch]["line"]
                if not argument_split:
                    argument_split = symbols[branches_branch]["split"]
                if not argument_end:
                    argument_end = symbols[branches_branch]["end"]

                symbols[branches_branch].update({"line": argument_line, "split": argument_split, "end": argument_end})

        if error == 0:
            if branch:
                raise NotExistingBranchError(branch)



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

"""br4nch - Data Structure Tree Builder for Python."""

__version__ = "1.1.2"
__author__ = "TRSTN4"
__author_email__ = "tristan@trstn4.com"
__description__ = "Data Structure Tree Generator for Python."
__url__ = "https://github.com/TRSTN4/br4nch"

import json
import urllib.request

import br4nch.create
import br4nch.duplicate
import br4nch.move
import br4nch.replace
import br4nch.delete
import br4nch.set
import br4nch.reset
import br4nch.load
import br4nch.export
import br4nch.display
import br4nch.utility


def check_version():
    """
    - Captures the response from the github api after which json loads the data and stores the value of the last
      released tag.
    - Prints a notification if the currently installed version is not equal to the latest version.
    """
    try:
        response = urllib.request.urlopen("https://api.github.com/repos/TRSTN4/br4nch/releases/latest")
    except urllib.error.URLError:
        return

    data = json.loads(response.read())
    latest_version = data["tag_name"]

    if __version__ != latest_version:
        print("[!] There is a new br4nch version (" + latest_version + ") available, The current installed br4nch "
              + "version (" + __version__ + ") is outdated.")


check_version()



# Copyright 2021-2022 by TRSTN4. All rights reserved.
# This file is part of the br4nch python package, and is released under the "GNU General Public License v3.0".
# Please see the LICENSE file that should have been included as part of this package.

from br4nch.utility.utility_librarian import branches, uids, paint_layer
from br4nch.utility.utility_handler import InstanceStringError, NotExistingBranchError
from br4nch.utility.utility_positioner import format_position


def arguments(branch, position):
    """
    - Gets the arguments and parses them to the 'DeleteLayer' class.
    """
    DeleteLayer(branch, position)


class DeleteLayer:
    def __init__(self, argument_branch, argument_position):
        """
        - Gets the arguments and parses them to the 'build_structure' function.
        """
        self.build_structure(argument_branch, argument_position)

    def build_structure(self, argument_branch, argument_position):
        """
        Lists:
          - If the given branch argument is not an instance of a list, then the branch argument will be set as a list.
          - If the given position argument is not an instance of a list, then the position argument will be set as a
            list.

        Operators:
          - If there a '*' in the 'argument_branch' list, Then it appends all existing branches to the 'argument_branch'
            list.

        Argument branch list loop:
          Errors:
            - If the branch value is not an instance of a string, then it raises an 'InstanceStringError' error.
            - If the branch is not in the 'branches' dictionary, it will throw a 'NotExistingBranchError' error.

          Branches list loop:
            Argument position list loop:
              - Calls the function 'get_layers' to get the needed values from the 'argument_position' variable.
              - Adds the returned value to the list 'queue_delete'.

          - Loops through the list 'queue_delete' and deletes all given layers from the dictionary 'branches'.
        """
        if not isinstance(argument_branch, list):
            argument_branch = [argument_branch]

        if not isinstance(argument_position, list):
            argument_position = [argument_position]

        if "*" in argument_branch:
            argument_branch.clear()
            for branches_branch in list(branches):
                argument_branch.append(branches_branch)

        for branch in argument_branch:
            error = 0

            if not isinstance(branch, str):
                raise InstanceStringError("branch", branch)

            for branches_branch in list(branches):
                if branch.lower() == branches_branch.lower():
                    error = error + 1

                    queue_delete = []

                    for position in format_position(branches_branch, argument_position.copy()):
                        value = self.get_layers(branches_branch, position,
                                                branches[branches_branch][list(branches[branches_branch])[0]])

                        queue_delete.append(value)

                    for delete_value in queue_delete:
                        if delete_value:
                            for layer, value in delete_value.items():
                                uids[branches_branch].remove(str(layer[-10:]))
                                paint_layer[branches_branch].pop(layer)

                                self.delete_layer_additions(branches_branch, value[layer])
                                del value[layer]

            if error == 0:
                if branch:
                    raise NotExistingBranchError(branch)

    def get_layers(self, branch, position, value):
        """
        Value dictionary loop:
          - For each value of the 'value' variable the 'count' variable is added with plus '1'.

          Count variable equal to the first value of 'position':
            If the length of the 'position' list is equal to '1':
              - Returns the current layer and previous value in an dictionary.

            - If the length of the 'position' list is not equal to '1' and there is a value of the 'value' variable,
              then the first value from the 'argument_move' list will be removed and the 'get_layers' function will be
              called again with the new value of the 'value' variable as argument.
        """
        count = 0
        previous_value = value

        for layer, value in value.items():
            count = count + 1

            if count == int(position[0]):
                if len(position) == 1:
                    return {layer: previous_value}
                else:
                    if value:
                        position.pop(0)
                        return self.get_layers(branch, position, value)

    def delete_layer_additions(self, branch, value):
        """
        Value dictionary loop:
          - Checks if the UID of the current value of 'layer' exists in the 'uids' list. If the UID is in the 'uids'
            list, then it will be removed from the list.
          - Checks if the 'layer' value exists in the 'paint_layer' branch list. If the layer is in the 'paint_layer'
            branch list, then the value of the current layer key in the 'paint_layer' dictionary will be removed.
          - If there is a value of the 'value' variable, the 'delete_layer_additions' function will be called again with
            the new value of the 'value' variable as argument.
        """
        for layer, value in value.items():
            if layer[-10:] in uids[branch]:
                uids[branch].remove(str(layer[-10:]))

            if layer in paint_layer[branch]:
                paint_layer[branch].pop(layer)

            if value:
                self.delete_layer_additions(branch, value)
